// 2009Aug17 JDM Added sawtooth
//
/*********************************************************************
*
* ANSI C Example program:
*    contGeneration.c
*
* Example Category:
*    AO
*
* Description:
*    This example demonstrates how to output a continuous periodic
*    waveform using an internal sample clock.
*
* Instructions for Running:
*    1. Select the Physical Channel to correspond to where your signal
*       is output on the DAQ device.
*    2. Enter the Minimum and Maximum Voltage Ranges.
*    3. Enter the desired rate for the generation. The onboard sample
*       clock will operate at this rate.
*    4. Select the desired waveform type.
*    5. The rest of the parameters in the Waveform Information section
*       will affect the way the waveform is created, before it's sent
*       to the analog output of the board. Select the number of samples
*       per cycle and the total number of cycles to be used as waveform
*       data.
*
* Steps:
*    1. Create a task.
*    2. Create an Analog Output Voltage channel.
*    3. Define the update Rate for the Voltage generation. Additionally,
*       define the sample mode to be continuous.
*    4. Write the waveform to the output buffer.
*    5. Call the Start function.
*    6. Loop continuously for 10 seconds. Check for errors every
*       100 ms using the IsTaskDone function.
*    7. Call the Clear Task function to clear the Task.
*    8. Display an error if any.
*
* I/O Connections Overview:
*    Make sure your signal output terminal matches the Physical Channel
*    I/O Control. For further connection information, refer to your
*    hardware reference manual.
*
* Recommended use:
*    Call Configure, Write and Start functions.
*    Call IsDone function in a loop.
*    Call Stop function at the end.
*
*********************************************************************/
#include "common.h"
#include <NIDAQmx.h>
#include <stdio.h>
#include <math.h>
#include <time.h>
#include <unistd.h>

#define PI		3.1415927
#define BUFFERSIZE	4096
#define NUMCHANNELS	4
#define NUMAOCHANNELS	3

#define DAQmxErrChk(functionCall) { if( DAQmxFailed(error=(functionCall)) ) { goto Error; } }


int modtrig(int command, float dither_freq, float dither_amplitude, float dither_smoothfrac,float dither_delay);
int TriangleWave(int number,float amplitude, float phase_offset, float64 *data);
int SawToothWave(int number,float amplitude, float phase_offset,  float64 *data);
int Trigger(int number, int width, uInt8 *data);
int BoxcarSmooth(int number, float64 *data, int width, float64 *data_sm);

TaskHandle  mod_taskHandle = 0, clock_taskHandle=0, trigger_taskHandle=0;


int modtrig(int command, float dither_freq, float dither_amplitude, float dither_smoothfrac, float dither_delay)
{
    int32       i,c; 
    char        errBuff[2048]={'\0'};
    int32	error = 0;

    // Channel parameters
    char        chan[] = "Dev1/ao0:3";
    float64     min = -10.0;
    float64     max = 10.0;

    // Timing parameters
    float64     sample_rate;

    // Data write parameters
    int32	data_size = BUFFERSIZE * NUMCHANNELS;
    float64     waveform[BUFFERSIZE], waveform1[BUFFERSIZE], waveform2[BUFFERSIZE];
    uInt8	ttl[BUFFERSIZE];
    float64	data[data_size];
    int32       pointsWritten;
    float64     timeout = 10.0;
    //FILE	*fp;  // NOT USED

    // Clock parameters
    float64 clock_freq;

    // Trigger parameters
    float	pulseWidth;
    float64	trigger_delay;
    float64 	dutyCycle;
    int32	lowTicks, highTicks;

    // Misc
    float       freq;
    int32	samples_per_chan=0;
    int32	smooth_width;


  switch (command) {

   case MODTRIG_START:

        /* get the frequency */
        freq=dither_freq;
	printf("Dither frequency: %0.1f Hz\n", freq);


        /* init the vectors */
        for (i=0; i<BUFFERSIZE; i++){
		waveform[i]=0;
		waveform1[i]=0;
		waveform2[i]=0;
	}
        for (i=0; i<data_size; i++){
		data[i]=0;
	}
	printf("Data size: %ld\n", data_size);


	/* set the number of samples per channel */
	samples_per_chan = BUFFERSIZE;
        printf("Number of samples per channel: %f\n", (float)samples_per_chan);

        /* make a triangle wave */
	/* JDM puts in an empirical phase offset factor, assuming 0.05 smoothfrac  */
	/* previous suggestion is 2*pi*dither_smoothfrac */

//        TriangleWave(samples_per_chan,
//		dither_amplitude, 
//		PI/2+.4,  // use dither_delay
//		waveform);
	SawToothWave(samples_per_chan,dither_amplitude,dither_delay,waveform);

        /* smooth it by SMOOTH_FRAC */
        smooth_width=(int)(samples_per_chan*dither_smoothfrac);
        printf("Smoothing boxcar width per cycle: %ld (%.0f%%)\n", 
		smooth_width, dither_smoothfrac*100.);
//        BoxcarSmooth(samples_per_chan,waveform,smooth_width,waveform1); // Triangle
// if SawTooth must smooth twice by half as much.
        BoxcarSmooth(samples_per_chan,waveform,smooth_width/2,waveform1);
        BoxcarSmooth(samples_per_chan,waveform1,smooth_width/2,waveform2);


        /* what sample rate (in samples/second) do we need */
	/* to get the freq we want? */
	sample_rate=freq*samples_per_chan;
        printf("Sample rate: %f samples/ms\n", sample_rate/1000);

	/* the TTL pulse */
	Trigger(samples_per_chan,
		sample_rate/1000,
		ttl);

        /* write the data (grouped by channel) for multiple channels */
        for (i=0;i<samples_per_chan;i++){
	   for (c=0;c<NUMAOCHANNELS;c++)
              data[i+c*samples_per_chan]=waveform2[i];
           data[i+3*samples_per_chan]=ttl[i];
	}

	/* write the data to a file for posterity */
//	fp = fopen("waveform.dat", "w");
//        for (i=0; i<data_size; i++){
//		fprintf(fp,"%ld %f\n",i,data[i]);
//	}
//	fclose(fp);



	/*
	 * MODULATION signal
	 */


	/* Create the analog out task */
	DAQmxErrChk(DAQmxCreateTask("modulation",&mod_taskHandle));


	/* Create the channel(s) and add them to the task */
	DAQmxErrChk(DAQmxCreateAOVoltageChan(
		mod_taskHandle,
		chan,
		"",
		min,
		max,
		DAQmx_Val_Volts,
		NULL
	));

	/* Set the source and rate of the clock and the number of samples */
	printf("Set the clock source and rate... ");
	DAQmxErrChk(DAQmxCfgSampClkTiming(
		mod_taskHandle,
		"Ctr0InternalOutput",
		sample_rate,
		DAQmx_Val_Rising,
		DAQmx_Val_ContSamps,
		samples_per_chan
	));
	printf("Done\n");

	/* Write the data to the buffer */
	printf("Writing waveform data to the buffer... ");
	DAQmxErrChk(DAQmxWriteAnalogF64(
		mod_taskHandle,
		samples_per_chan,
		0,			// auto-start set to false
		timeout,
		DAQmx_Val_GroupByChannel,
		data,
		&pointsWritten,
		NULL
	));
	printf("Done\n");
    	printf("Number of points written per channel:  %ld\n", pointsWritten);

	/* sync the the mod signal with the trigger */
	DAQmxErrChk(DAQmxCfgDigEdgeStartTrig(
	   mod_taskHandle, 
	   "/Dev1/Ctr0InternalOutput", 
	   DAQmx_Val_Rising
	));

	/* start the task */
	DAQmxErrChk(DAQmxStartTask(mod_taskHandle));


	/*
	 * CLOCK
	 */

	clock_freq = freq*samples_per_chan;
	printf("Clock frequency: %f Hz\n", clock_freq);

	/* create the clock task */
	DAQmxErrChk(DAQmxCreateTask("clock",&clock_taskHandle));

	/* setup the clock */
	DAQmxErrChk(DAQmxCreateCOPulseChanFreq(
		clock_taskHandle,
		"Dev1/ctr0",
		"",
		DAQmx_Val_Hz,
		DAQmx_Val_Low,
		0.0,
		clock_freq,
		0.5
	));

	/* implicit timing */
	DAQmxErrChk(DAQmxCfgImplicitTiming(
	   clock_taskHandle,
	   DAQmx_Val_ContSamps,
	   0
	));



	/*
	 * TRIGGER
	 */ 
	pulseWidth=0.5; //in milliseconds
        if ((dutyCycle = (pulseWidth/1000.) * clock_freq) < 2) {
           dutyCycle=2;
           printf("WARNING:  dutyCycle too low, setting to %f (%f ms)\n",
                   dutyCycle, dutyCycle*1000/clock_freq);
        } else {
           printf("dutyCycle: %f\n", dutyCycle);
        }


	/* phase offset between trigger and modulation signal */
	trigger_delay = samples_per_chan/4.-smooth_width/2;

	highTicks=dutyCycle;
	lowTicks=samples_per_chan/2-highTicks;

	/* create the task */
	DAQmxErrChk(DAQmxCreateTask("trigger",&trigger_taskHandle));

	/* setup the trigger in terms of clock ticks */
	DAQmxErrChk(DAQmxCreateCOPulseChanTicks(
	    trigger_taskHandle, 
	    "Dev1/ctr1", 
	    "", 
	    "Ctr0InternalOutput", 
	    DAQmx_Val_Low, 
	    //trigger_delay, 
	    0, 
	    lowTicks,
	    highTicks
	));

	DAQmxErrChk(DAQmxCfgImplicitTiming(
	   trigger_taskHandle,
	   DAQmx_Val_ContSamps,
	   0
	));

	/* start the trigger when the clock starts */
	DAQmxErrChk(DAQmxCfgDigEdgeStartTrig (
	   trigger_taskHandle, 
	   "Ctr0InternalOutput",
	   DAQmx_Val_Rising
	));

	DAQmxErrChk(DAQmxStartTask(trigger_taskHandle));


	/* start the clock, which starts the mod and trig */
	DAQmxErrChk(DAQmxStartTask(clock_taskHandle));

	printf("Done with start trigger\n");

   break; /* end START case*/


   case MODTRIG_STOP:

	/* if the task has let go, stop and clear it */
	if(mod_taskHandle!=0){
	   printf("Stopping and Clearing Mod Task... ");
	   DAQmxErrChk(DAQmxStopTask(mod_taskHandle));
	   DAQmxErrChk(DAQmxClearTask(mod_taskHandle));
	   printf("Done\n");
	}

	if(trigger_taskHandle!=0){
	   printf("Stopping and Clearing Trigger Task... ");
	   DAQmxErrChk(DAQmxStopTask(trigger_taskHandle));
	   DAQmxErrChk(DAQmxClearTask(trigger_taskHandle));
	   printf("Done\n");
	}

	if(clock_taskHandle!=0){
	   printf("Stopping and Clearing clock Task... ");
	   DAQmxErrChk(DAQmxStopTask(clock_taskHandle));
	   DAQmxErrChk(DAQmxClearTask(clock_taskHandle));
	   printf("Done\n");
	}

   break; /* end STOP case*/


   case MODTRIG_RESET:

	/* Reset the device */
	printf("Resetting NI device... ");
	DAQmxErrChk(DAQmxResetDevice("Dev1"));
	printf("Done\n");

   break; /* end RESET case*/


  } /* end switch */


Error:
        /* if there was an error, get a more user-friendly message */
        if( DAQmxFailed(error) ){
            printf("\n\nERROR ERROR ERROR\n\n");
            DAQmxGetExtendedErrorInfo(errBuff,2048);
        }

        /* report the error if there was one */
        if( DAQmxFailed(error) )
            printf("DAQmx Error: %s\n",errBuff);


	return 0;
}


/* function to create a triangle waveform */
int TriangleWave(int number, float amplitude, float phase_offset, float64 *data)
{
        int i;
        float64 phase;

        for (i=0;i<number;i++)
        {
                phase=2*PI*(float)i/(float)number + phase_offset;
		// data amp is centered on 5 (middle of DAQ range
		*data++ = amplitude*(asin(sin(phase))/PI)+5.;
        }

        return 0;
}
/* function to create a triangle waveform */
int SawToothWave(int number, float amplitude, float offset, float64 *data)
{
// offset is in fractions of sawtooth
        int i;
        float64 phase;
	float64 intpart;

        for (i=0;i<number;i++)
        {
                phase=2*(float)i/(float)number + offset;
                // data amp is centered on 5 (middle of DAQ range
                *data++ = amplitude*(modf(phase,&intpart)-.5)+5.;
//		printf("SawTooth bin, amp:%f6.2 %f6.2\n",phase,amplitude*(modf(phase,&intpart)-.5)+5.);
        }

        return 0;
}


/* function to create a triangle waveform */
int Trigger(int number, int width, uInt8 *data)
{
        int i;

        for (i=0;i<number;i++)
        {
		if ((i>0) && (i<width)){
			*data++ = 1;
		} else {
			*data++ = 0;
		}
        }

        return 0;
}

/* boxcar smoothes any waveform */
// JDM will smooth around cycliclally [worked ok before too!]
int BoxcarSmooth(int number, float64 *data, int width, float64 *data_sm)
{
        float64 sum;
        int i,j,index;

        for (i=0;i<number;i++)
        {
                sum=0.0;
                for (j=0;j<width-1;j++)
                {
                        index=i+j-width/2;
                        if (index >= number) index=index-number;
                        if (index < 0) index=index+number;
                        sum+=data[index];
                }
                sum/=width;
                *data_sm++=sum;
        }

        return 0;
}

